from __future__ import print_function

import os
import platform
import string
import subprocess
import sys
import threading

class RunCmdResult(object):
    """ Container for the results of run_cmd """
    def __init__(self,echoStdout,echoStderr):
        self.retCode = 0;
        self.stdout = [];
        self.stderr = [];
        self.echoStdout = echoStdout;
        self.echoStderr = echoStderr;
    def addStdOut(self,lines):
        if (len(lines)==0):
            return;
        lines = map(lambda line: line.rstrip('\n'),lines)
        if (self.echoStdout):
            print("\n".join(lines), file=sys.stdout)
        self.stdout.extend(lines)
    def addStderr(self,lines):
        if (len(lines)==0):
            return;
        lines = map(lambda line: line.rstrip('\n'),lines)
        if (self.echoStderr):
            print("\n".join(lines), file=sys.stderr)
        self.stderr.extend(lines)

def run(command,echoCommand=True):
    """ Run a command, but do not capture its output. Raise exception on non-zero status """
    if echoCommand:
        print( command )
        sys.stdout.flush()
    result = os.system(command)
    if 0 != result:
        raise Exception("Command '{}' failed with status {}".format(command,result))
        
def run_cmd(args,throwOnNonZero = True,echoCommand=True,echoStdout=False,echoStderr=True):
    """ Run a command and capture its output. Optionally raise exception on non-zero status """
    if echoCommand:
        print(' '.join(args))
        sys.stdout.flush()
    # set the use show window flag, might make conditional on being in Windows:
    if platform.system() == 'Windows':
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    else:
        startupinfo = None
    # pass as the startupinfo keyword argument:
    p=subprocess.Popen(args,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             stdin=open(os.devnull), # subprocess.PIPE,
                             startupinfo=startupinfo)

    result = RunCmdResult(echoStdout,echoStderr);

    class StdoutReaderThread(threading.Thread):
        def __init__(self, stream, result):
            threading.Thread.__init__(self)
            self.stream = stream
            self.result = result
        def run(self):
            while True:
                lines = self.stream.readlines(1024)
                if len(lines) == 0:
                    break
                self.result.addStdOut(lines)
    class StderrReaderThread(threading.Thread):
        def __init__(self, stream, result):
            threading.Thread.__init__(self)
            self.stream = stream
            self.result = result
        def run(self):
            while True:
                lines = self.stream.readlines(1024)
                if len(lines) == 0:
                    break
                self.result.addStderr(lines)

    stdOutReader = StdoutReaderThread(p.stdout,result)
    stdOutReader.start()
    stdErrReader = StderrReaderThread(p.stderr,result)
    stdErrReader.start()

    result.retCode = p.wait()
    stdOutReader.join()
    stdErrReader.join()

    if (throwOnNonZero and result.retCode != 0):
        sys.stdout.flush();
        print( result.stderr, file=sys.stderr)
        sys.stderr.flush();
        raise Exception("Command Exited with status=%d" %result.retCode)
    return result;




import os
import pyperclip

from AwsProcessor import AwsProcessor
from awsHelpers.AwsConnectionFactory import AwsConnectionFactory
from CommandArgumentParser import CommandArgumentParser
from stdplus import *

from pprint import pprint

class WrappedEvent:
    def __init__(self,event):
        # pprint(event)
        self.event = event
        self.logical_id = event.id
        self.resource_status = event.logical_resource_id + ":" + event.resource_status
        self.resource_status_reason = event.resource_status_reason

class WrappedOutput:
    def __init__(self,output):
        # pprint(output)
        self.output = output
        self.logical_id = output['OutputKey']
        self.resource_status = output['OutputValue']
        self.resource_status_reason = defaultifyDict(output,'Description','')
        
class AwsStack(AwsProcessor):
    def __init__(self,stack,logicalName,parent):
        """Construct an AwsStack command processor"""
        AwsProcessor.__init__(self,parent.raw_prompt + "/stack:" + logicalName,parent)
        self.wrappedStack = self.wrapStack(stack)
        self.printStack(self.wrappedStack)

    def wrapStackEvents(self,stack):
        events = {}
        i = 0;
        if None != stack.events:
            for event in stack.events.all():
                events[i] = WrappedEvent(event)
                i = i + 1
        return events;

    def wrapStackOutputs(self,stack):
        outputs = {}
        i = 0;
        if None != stack.outputs:
            for output in stack.outputs:
                outputs[i] = WrappedOutput(output)
                i = i + 1
        return outputs

    def wrapStack(self,stack):
        result = {};
        result['rawStack'] = stack;

        resourcesByType = {};
        for resource in stack.resource_summaries.all():
            if not resource.resource_type in resourcesByType:
                resourcesByType[resource.resource_type] = {}
            resourcesByType[resource.resource_type][resource.logical_id] = resource;

        resourcesByType['events'] = self.wrapStackEvents(stack)
        resourcesByType['outputs'] = self.wrapStackOutputs(stack)

        result['resourcesByTypeName'] = resourcesByType;

        resourcesByTypeIndex = {};
        for resourceType, resources in resourcesByType.items():
            resourcesByTypeIndex[resourceType] = {};
            index = 0
            for name,resource in resources.items():
                resourcesByTypeIndex[resourceType][index] = resource
                index += 1
        result['resourcesByTypeIndex'] = resourcesByTypeIndex;
        return result
        
    def printStack(self,wrappedStack,include=None,filters=["*"]):
        """Prints the stack"""
        rawStack = wrappedStack['rawStack']
        print "==== Stack {} ====".format(rawStack.name)
        print "Status: {} {}".format(rawStack.stack_status,defaultify(rawStack.stack_status_reason,''))

        for resourceType, resources in wrappedStack['resourcesByTypeIndex'].items():
            if resourceType in AwsProcessor.resourceTypeAliases:
                resourceType = AwsProcessor.resourceTypeAliases[resourceType];
            if (None == include or resourceType in include) and len(resources):
                print "== {}:".format(resourceType)
                logicalIdWidth = 1
                resourceStatusWidth = 1
                resourceStatusReasonWidth = 1
                for index, resource in resources.items():
                    logicalIdWidth = max(logicalIdWidth,len(resource.logical_id))
                    resourceStatusWidth = max(resourceStatusWidth,len(resource.resource_status))
                    resourceStatusReasonWidth = max(resourceStatusReasonWidth,len(defaultify(resource.resource_status_reason,'')))
                frm = "    {{0:3d}}: {{1:{0}}} {{2:{1}}} {{3}}".format(logicalIdWidth,resourceStatusWidth)
                for index, resource in resources.items():
                    if fnmatches(resource.logical_id.lower(),filters):
                        print frm.format(index,resource.logical_id,resource.resource_status,defaultify(resource.resource_status_reason,''))

    def do_browse(self,args):
        """Open the current stack in a browser."""
        rawStack = self.wrappedStack['rawStack']
        os.system("open -a \"Google Chrome\" https://us-west-2.console.aws.amazon.com/cloudformation/home?region=us-west-2#/stack/detail?stackId={}".format(rawStack.stack_id))
                
    def do_refresh(self,args):
        """Refresh view of the current stack. refresh -h for detailed help"""
        self.wrappedStack = self.wrapStack(AwsConnectionFactory.instance.getCfResource().Stack(self.wrappedStack['rawStack'].name))
        
    def do_print(self,args):
        """Print the current stack. print -h for detailed help"""
        parser = CommandArgumentParser("print")
        parser.add_argument('-r','--refresh',dest='refresh',action='store_true',help='refresh view of the current stack')
        parser.add_argument('-i','--include',dest='include',default=None,nargs='+',help='resource types to include')
        parser.add_argument(dest='filters',nargs='*',default=["*"],help='Filter stacks');
        args = vars(parser.parse_args(args))

        if args['refresh']:
            self.do_refresh('')

        self.printStack(self.wrappedStack,args['include'],args['filters'])
        
    def do_resource(self,args):
        """Go to the specified resource. resource -h for detailed help"""
        parser = CommandArgumentParser("resource")
        parser.add_argument('-i','--logical-id',dest='logical-id',help='logical id of the child resource');
        args = vars(parser.parse_args(args))

        stackName = self.wrappedStack['rawStack'].name
        logicalId = args['logical-id']
        self.stackResource(stackName,logicalId)

    def do_asg(self,args):
        """Go to the specified auto scaling group. asg -h for detailed help"""
        parser = CommandArgumentParser("asg")
        parser.add_argument(dest='asg',help='asg index or name');
        args = vars(parser.parse_args(args))

        print "loading auto scaling group {}".format(args['asg'])
        try:
            index = int(args['asg'])
            asgSummary = self.wrappedStack['resourcesByTypeIndex']['AWS::AutoScaling::AutoScalingGroup'][index]
        except:
            asgSummary = self.wrappedStack['resourcesByTypeName']['AWS::AutoScaling::AutoScalingGroup'][args['asg']]

        self.stackResource(asgSummary.stack_name,asgSummary.logical_id)

    def do_eni(self,args):
        """Go to the specified eni. eni -h for detailed help."""
        parser = CommandArgumentParser("eni")
        parser.add_argument(dest='eni',help='eni index or name');
        args = vars(parser.parse_args(args))

        print "loading eni {}".format(args['eni'])
        try:
            index = int(args['eni'])
            eniSummary = self.wrappedStack['resourcesByTypeIndex']['AWS::EC2::NetworkInterface'][index]
        except ValueError:
            eniSummary = self.wrappedStack['resourcesByTypeName']['AWS::EC2::NetworkInterface'][args['eni']]

        pprint(eniSummary)
        self.stackResource(eniSummary.stack_name,eniSummary.logical_id)

    def do_stack(self,args):
        """Go to the specified stack. stack -h for detailed help."""
        parser = CommandArgumentParser("stack")
        parser.add_argument(dest='stack',help='stack index or name');
        args = vars(parser.parse_args(args))

        print "loading stack {}".format(args['stack'])
        try:
            index = int(args['stack'])            
            stackSummary = self.wrappedStack['resourcesByTypeIndex']['AWS::CloudFormation::Stack'][index]
        except ValueError:
            stackSummary = self.wrappedStack['resourcesByTypeName']['AWS::CloudFormation::Stack'][args['stack']]

        self.stackResource(stackSummary.stack_name,stackSummary.logical_id)

    def do_template(self,args):
        """Print the template for the current stack. template -h for detailed help"""
        parser = CommandArgumentParser("template")
        args = vars(parser.parse_args(args))

        print "reading template for stack."
        rawStack = self.wrappedStack['rawStack']
        template = AwsConnectionFactory.getCfClient().get_template(StackName=rawStack.name)
        print template['TemplateBody']

    def getWrappedItem(self,typeName,IdOrName):
        try:
            index = int(IdOrName)
            return self.wrappedStack['resourcesByTypeIndex'][typeName][index]
        except ValueError:
            return self.wrappedStack['resourcesByTypeName'][typeName][IdOrName]
        
    def getOutputs(self,outputs):
        values = []        
        for output in outputs:
            values.append(self.getWrappedItem('outputs',output).resource_status)
        pprint(values)
        return values

    def do_copy(self,args):
        """Copy specified id to stack. copy -h for detailed help."""
        parser = CommandArgumentParser("copy")
        parser.add_argument('-a','--asg',dest='asg',nargs='+',required=False,default=[],help='Copy specified ASG info.')
        parser.add_argument('-o','--output',dest='output',nargs='+',required=False,default=[],help='Copy specified output info.')        
        args = vars(parser.parse_args(args))
        if args['output']:
            pyperclip.copy("\n".join(self.getOutputs(args['output'])))

    def do_stacks(self,args):
        """Same as print -r --include stack"""
        self.do_print(args + " -r --include stack" )




import yaml

def describeInstances(ec2Client,InstanceIds):
    response = ec2Client.describe_instances(InstanceIds=InstanceIds)
    instanceDescriptions = response['Reservations']
    instances = {}
    for instance in instanceDescriptions:
        instance = instance['Instances'][0]
        instanceId = instance['InstanceId']
        instances[instanceId]=instance
    return instances
    
def instanceIdsToPrivateIps(ec2Client,instanceIds):
    if instanceIds:
        instances = describeInstances(ec2Client,instanceIds)
        result = {}
        for instanceId,instance in instances.items():
            result[instanceId] = []
            for networkInterface in instance['NetworkInterfaces']:
                privateIp = networkInterface['PrivateIpAddress']
                result[instanceId].append(privateIp)
        return result
    else:
        return {}



from _readfile import readfile
from _sshConfig import *
from _writefile import writefile
from _fexecvp import fexecvp
from _fnmatches import fnmatches
from _run_cmd import *
from _defaultify import *
from _isInt import *

import re

__all__ = [
    'defaultify',
    'defaultify',
    'defaultifyDict',
    'defaultifyDict',
    'fexecvp',
    'fnmatches',
    'getSshHost',
    'isInt',
    'isIp',
    'keyscanHost',
    'readSshConfig',
    'readfile',
    'removeKnownHosts',
    'resetKnownHost',
    'resetKnownHosts',
    'run',
    'run_cmd',
    'writefile'
]

def contains(small, big):
    for i in xrange(len(big)-len(small)+1):
        for j in xrange(len(small)):
            if big[i+j] != small[j]:
                break
        else:
            return i, i+len(small)
    return False

def isIp(string):
    return None != re.match("^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$",string)



def getInstanceIds(asgClient,asgName):
    asgDescription = asgClient.describe_auto_scaling_groups(AutoScalingGroupNames=[asgName])
    if asgDescription['AutoScalingGroups'] and asgDescription['AutoScalingGroups'][0]:
        instances = asgDescription['AutoScalingGroups'][0]['Instances']
        instanceIds = []
        for instance in instances:
            instanceIds.add(instance['InstanceId'])
        return instanceIds
    else:
        return []    



def defaultify(value,default):
    if None == value:
        return default
    else:
        return value

def defaultifyDict(dictionary,key,default):
    if key in dictionary:
        return defaultify(dictionary[key],default)
    else:
        return default



from AwsProcessor import AwsProcessor
from awsHelpers.AwsConnectionFactory import AwsConnectionFactory

from pprint import pprint

class AwsEni(AwsProcessor):
    def __init__(self,physicalId,parent):
        """Construct an AwsEni command processor"""
        AwsProcessor.__init__(self,parent.raw_prompt + "/eni:" + physicalId,parent)
        self.physicalId = physicalId
        self.do_refresh('')

    def do_refresh(self,args):
        """Refresh the view of the eni"""
        pprint(AwsConnectionFactory.getEc2Client().describe_network_interfaces(NetworkInterfaceIds=[self.physicalId]));



import os
import re
import fnmatch
import threading

from pprint import pprint
from stdplus import readfile
from _run_cmd import run

knownHostsLock = threading.Lock()

class SshConfigBlock(list):
    def __init__(self,line,lineNumber):
        list.__init__(self)
        self.line = line
        self.lineNumber = lineNumber

    def __str__(self):
        return "{}{}{}".format(self.line,"\n" if len(self) > 0 else "","\n".join(x.__repr__() for x in self) )

    def __repr__(self):
        return "{}:<{}>{}{}".format(self.lineNumber,self.line,"\n" if len(self) > 0 else "","\n".join(x.__repr__() for x in self) )

class SshConfig(SshConfigBlock):
    def __init__(self):
        SshConfigBlock.__init__(self,'',0)
        self.hosts = {}
    def __str__(self):
        return "\n".join(x.__repr__() for x in self)
    def __repr__(self):
        return "\n".join(x.__repr__() for x in self)

class WithSettings(SshConfigBlock):
    def __init__(self,line,lineNumber):
        SshConfigBlock.__init__(self,line,lineNumber)
        self.settings = {}
    def addSetting(self,setting):
        if not setting.name in self.settings:
            self.settings[setting.name] = []
        self.settings[setting.name].append(setting)
        
class Host(WithSettings):
    def __init__(self,line,lineNumber):
        WithSettings.__init__(self,line,lineNumber)
        m = re.search('^.*Host[ \t]+([^#]*)(#.*)*',line)
        self.name = m.group(1)

class Comment(SshConfigBlock):
    pass

class Match(WithSettings):
    pass

class Setting(SshConfigBlock):
    def __init__(self,line,lineNumber):
        SshConfigBlock.__init__(self,line,lineNumber)
        m = re.search('^[ \t]*([A-Za-z_]*)[ \t]+([^#]*)(#.*)*',line)
        self.name = m.group(1)
        self.value = m.group(2)
        self.comment = m.group(3)

def parseSshConfig(contents):
    lines = contents.split('\n')
    lineNumber = 0
    rootBlock = SshConfig()
    currentBlock = rootBlock
    for line in lines:
        lineNumber += 1
        if 0 == len(line.strip()) or line.startswith('#'):
            currentBlock.append(Comment(line,lineNumber))
        elif line.strip().startswith('Host '):
            currentBlock = Host(line,lineNumber)
            rootBlock.append(currentBlock)
            rootBlock.hosts[currentBlock.name] = currentBlock
        elif line.strip().startswith('Match '):
            currentBlock = Match(line,lineNumber)
            rootBlock.append(currentBlock)
        else:
            setting = Setting(line,lineNumber)
            currentBlock.append(setting)
            if isinstance(currentBlock,Host) or isinstance(currentBlock,Match):
                currentBlock.addSetting(setting)
            
    return rootBlock

def readSshConfig(path=os.path.expanduser('~/.ssh/config')):
    config = readfile(path)
    return parseSshConfig(config)

def getSshHost(sshConfig,host):
    for pattern,settings in sshConfig.hosts.iteritems():
        if fnmatch.fnmatch(host,pattern):
            return settings
        
def resetKnownHost(ip):
    removeKnownHosts([ip])
    keyscanHost(ip)

def resetKnownHosts(ips):
    removeKnownHosts(ips)
    for ip in ips:
        keyscanHost(ip)

def keyscanHost(ip):
    sshConfig = readSshConfig()
    sshHost = getSshHost(sshConfig,ip)
    proxyCommand = ''
    if sshHost and 'ProxyCommand' in sshHost.settings and sshHost.settings['ProxyCommand'] and sshHost.settings['ProxyCommand'][0].value:
        proxyCommand = sshHost.settings['ProxyCommand'][0].value.replace("-W %h:%p",'')
    global knownHostsLock
    with knownHostsLock:
        run("{} ssh-keyscan -H {} >> ~/.ssh/known_hosts".format(proxyCommand, ip))    

def removeKnownHosts(ips):
    if not ips:
        return
    args = []
    for ip in ips:
        args.append("-R {}".format(ip))
    global knownHostsLock
    with knownHostsLock:
        run("ssh-keygen {}".format(" ".join(args)))




import boto3
import json
import os

from botocore import configloader
from pprint import pprint
from stdplus import readfile
from stdplus import writefile

aws_config_dir = os.path.join(os.path.expanduser("~"), ".aws")

class AwsConnectionFactory:
    instance = None
    
    def __init__(self,credentials=None,profile='default'):
        if None == credentials:
            credentialsFilename = self.getAwsMfaCredentialsFilename(profile)
            try:
                credentials = json.loads(readfile(credentialsFilename))['Credentials']
            except IOError:
                # print "WARN: IOError reading credentials file:{}".format(credentialsFilename)
                pass
            except ValueError:
                # print "WARN: ValueError reading credentials file:{}".format(credentialsFilename)
                pass
        self.setMfaCredentials(credentials,profile)

    @staticmethod
    def resetInstance(credentials=None,profile='default'):
        AwsConnectionFactory.instance = AwsConnectionFactory(credentials,profile)

    def getAwsMfaCredentialsFilename(self,profile='default'):
        credentials_file_name = 'mfa_credentials'
        if not 'default' == profile:
            credentials_file_name = "{}_{}".format(profile,'mfa_credentials')

        credentials_file  = os.path.join(aws_config_dir, credentials_file_name)
        return credentials_file

    def getAwsCredentialsFilename(self):
        return os.path.join(aws_config_dir,'credentials')

    def load_arn(self,profile):
        arn_file_name = 'mfa_device'
        if not profile == 'default':
            arn_file_name = "{}_{}".format(profile,arn_file_name)

        arn_file = os.path.join(aws_config_dir, arn_file_name)

        print "arn_file:{} [profile:{}]".format(arn_file,profile)
        if os.access(arn_file,os.R_OK):
            return readfile(arn_file).strip()
        else:
            arn = self.load_arn_from_aws(profile)
            writefile(arn_file, arn)
            return arn
    
    def _strConfig(self,config):
        result = ""
        for section,settings in config.items():
            result += "[{}]\n".format(section)
            for key,value in settings.items():
                result += "{} = {}\n".format(key,value)
            result += "\n"
        return result

    def storeAwsMfa(self,credentials,profile):
        if not None == credentials:
            writefile(self.getAwsMfaCredentialsFilename(profile),json.dumps({'Credentials':credentials}))

    def storeAwsCredentials(self,credentials,profile):
        if credentials == None:
            return
        awsCredentials = configloader.raw_config_parse(self.getAwsCredentialsFilename(),False)
        if not profile in awsCredentials:
            awsCredentials[profile]={}
        awsCredentials[profile]['aws_access_key_id'] = credentials['AccessKeyId']
        awsCredentials[profile]['aws_secret_access_key'] = credentials['SecretAccessKey']
        if 'SessionToken' in credentials:
            awsCredentials[profile]['aws_session_token'] = credentials['SessionToken']
        awsCredentials = self._strConfig(awsCredentials)
        writefile(self.getAwsCredentialsFilename(),awsCredentials)
    
    def setMfaCredentials(self,credentials,profile='default'):
        self.storeAwsMfa(credentials,profile)
        self.storeAwsCredentials(credentials,profile + "-mfa");
        
        self.credentials = credentials
        self.session = None
        self.profile = profile

    def getSession(self):
        if self.session == None:
            if self.credentials == None:
                # print("Getting session w/ credentials from env")
                self.session = boto3.session.Session()
            else:
                # print("Getting session w/ credentials:")
                # pprint(self.credentials)
                self.session = boto3.session.Session(aws_access_key_id=self.credentials['AccessKeyId'],
                                                     aws_secret_access_key=self.credentials['SecretAccessKey'],
                                                     aws_session_token=self.credentials['SessionToken'])
                # print(self.session)
        # print "Session obtained"
        return self.session

    @staticmethod
    def getAsgClient():
        return AwsConnectionFactory.instance._getAsgClient()
    
    def _getAsgClient(self):
        return self.getSession().client('autoscaling')

    def getAsgResource(self):
        return self.getSession().resource('autoscaling')
    
    def getCfResource(self):
        return self.getSession().resource('cloudformation')

    @staticmethod
    def getCfClient():
        return AwsConnectionFactory.instance._getCfClient()

    def _getCfClient(self):
        return self.getSession().client('cloudformation')

    @staticmethod
    def getEc2Client():
        return AwsConnectionFactory.instance._getEc2Client()
    
    def _getEc2Client(self):
        return self.getSession().client('ec2')

    def getEc2Resource(self):
        return self.getSession().resource('ec2')

    def getS3Client(self):
        return self.getSession().client('s3')

    def getS3Resource(self):
        return self.getSession().resource('s3')
    
    def getProfile(self):
        return self.profile

AwsConnectionFactory.instance = AwsConnectionFactory()
